@@ -579,19 +579,17 @@ void telephony_nr_and_ec_req(void *telephony_device, gboolean enable)
 
 void telephony_key_press_req(void *telephony_device, const char *keys)
 {
-	struct voice_call *active, *waiting;
+	struct voice_call *active, *incoming;
 	int err;
 
 	DBG("telephony-ofono: got key press request for %s", keys);
 
-	waiting = find_vc_with_status(CALL_STATUS_INCOMING);
-	if (!waiting)
-		waiting = find_vc_with_status(CALL_STATUS_DIALING);
+	incoming = find_vc_with_status(CALL_STATUS_INCOMING);
 
 	active = find_vc_with_status(CALL_STATUS_ACTIVE);
 
-	if (waiting)
-		err = answer_call(waiting);
+	if (incoming)
+		err = answer_call(incoming);
 	else if (active)
 		err = release_call(active);
 	else
@@ -935,6 +933,15 @@ static int parse_network_properties(DBusMessageIter *properties)
 				AG_FEATURE_ENHANCED_CALL_CONTROL |
 				AG_FEATURE_EXTENDED_ERROR_RESULT_CODES |
 				AG_FEATURE_THREE_WAY_CALLING;
+	int i;
+
+	/* Reset indicators */
+	for (i = 0; ofono_indicators[i].desc != NULL; i++) {
+		if (g_str_equal(ofono_indicators[i].desc, "battchg"))
+			ofono_indicators[i].val = 5;
+		else
+			ofono_indicators[i].val = 0;
+	}
 
 	while (dbus_message_iter_get_arg_type(properties)
 						== DBUS_TYPE_DICT_ENTRY) {
@@ -1033,6 +1040,8 @@ static void modem_removed(const char *path)
 
 	g_free(net.operator_name);
 	net.operator_name = NULL;
+	net.status = NETWORK_REG_STATUS_NOSERV;
+	net.signals_bar = 0;
 
 	g_free(modem_obj_path);
 	modem_obj_path = NULL;
@@ -1105,7 +1114,7 @@ static void get_modems_reply(DBusPendingCall *call, void *user_data)
 	DBusMessage *reply;
 	DBusMessageIter iter, entry;
 
-	DBG("list_modem_reply is called\n");
+	DBG("");
 	reply = dbus_pending_call_steal_reply(call);
 
 	dbus_error_init(&err);
