@@ -510,14 +510,23 @@ static void hdp_mdl_reconn_cb(struct mcap_mdl *mdl, GError *err, gpointer data)
 	}
 
 	fd = mcap_mdl_get_fd(dc_data->hdp_chann->mdl);
-	if (fd < 0)
+	if (fd < 0) {
 		reply = g_dbus_create_error(dc_data->msg,
 						ERROR_INTERFACE ".HealthError",
 						"Cannot get file descriptor");
+		g_dbus_send_message(dc_data->conn, reply);
+		return;
+	}
 
-	reply = g_dbus_create_reply(dc_data->msg, DBUS_TYPE_UNIX_FD, &fd,
-							DBUS_TYPE_INVALID);
+	reply = g_dbus_create_reply(dc_data->msg, DBUS_TYPE_UNIX_FD,
+							&fd, DBUS_TYPE_INVALID);
 	g_dbus_send_message(dc_data->conn, reply);
+
+	g_dbus_emit_signal(dc_data->conn,
+			device_get_path(dc_data->hdp_chann->dev->dev),
+			HEALTH_DEVICE, "ChannelConnected",
+			DBUS_TYPE_OBJECT_PATH, &dc_data->hdp_chann->path,
+			DBUS_TYPE_INVALID);
 }
 
 static void hdp_get_dcpsm_cb(uint16_t dcpsm, gpointer user_data, GError *err)
@@ -874,6 +883,7 @@ static gboolean check_channel_conf(struct hdp_channel *chan)
 	GError *err = NULL;
 	GIOChannel *io;
 	uint8_t mode;
+	uint16_t imtu, omtu;
 	int fd;
 
 	fd = mcap_mdl_get_fd(chan->mdl);
@@ -883,6 +893,8 @@ static gboolean check_channel_conf(struct hdp_channel *chan)
 
 	if (!bt_io_get(io, BT_IO_L2CAP, &err,
 			BT_IO_OPT_MODE, &mode,
+			BT_IO_OPT_IMTU, &imtu,
+			BT_IO_OPT_OMTU, &omtu,
 			BT_IO_OPT_INVALID)) {
 		error("Error: %s", err->message);
 		g_io_channel_unref(io);
@@ -894,13 +906,30 @@ static gboolean check_channel_conf(struct hdp_channel *chan)
 
 	switch (chan->config) {
 	case HDP_RELIABLE_DC:
-		return mode == L2CAP_MODE_ERTM;
+		if (mode != L2CAP_MODE_ERTM)
+			return FALSE;
+		break;
 	case HDP_STREAMING_DC:
-		return mode == L2CAP_MODE_STREAMING;
+		if (mode != L2CAP_MODE_STREAMING)
+			return FALSE;
+		break;
 	default:
 		error("Error: Connected with unknown configuration");
 		return FALSE;
 	}
+
+	DBG("MDL imtu %d omtu %d Channel imtu %d omtu %d", imtu, omtu,
+						chan->imtu, chan->omtu);
+
+	if (!chan->imtu)
+		chan->imtu = imtu;
+	if (!chan->omtu)
+		chan->omtu = omtu;
+
+	if (chan->imtu != imtu || chan->omtu != omtu)
+		return FALSE;
+
+	return TRUE;
 }
 
 static void hdp_mcap_mdl_connected_cb(struct mcap_mdl *mdl, void *data)
@@ -1309,7 +1338,7 @@ static gboolean update_adapter(struct hdp_adapter *hdp_adapter)
 		goto update;
 
 	adapter_get_address(hdp_adapter->btd_adapter, &addr);
-	hdp_adapter->mi = mcap_create_instance(&addr, BT_IO_SEC_HIGH, 0, 0,
+	hdp_adapter->mi = mcap_create_instance(&addr, BT_IO_SEC_MEDIUM, 0, 0,
 					mcl_connected, mcl_reconnected,
 					mcl_disconnected, mcl_uncached,
 					NULL, /* CSP is not used by now */
