@@ -54,6 +54,7 @@
 struct connect_mcl {
 	struct mcap_mcl		*mcl;		/* MCL for this operation */
 	mcap_mcl_connect_cb	connect_cb;	/* Connect callback */
+	GDestroyNotify		destroy;	/* Destroy callback */
 	gpointer		user_data;	/* Callback user data */
 };
 
@@ -66,6 +67,7 @@ typedef union {
 struct mcap_mdl_op_cb {
 	struct mcap_mdl		*mdl;		/* MDL for this operation */
 	mcap_cb_type		cb;		/* Operation callback */
+	GDestroyNotify		destroy;	/* Destroy callback */
 	gpointer		user_data;	/* Callback user data */
 };
 
@@ -261,6 +263,23 @@ static gint cmp_mdl_state(gconstpointer a, gconstpointer b)
 		return 1;
 }
 
+static void free_mcap_mdl_op(struct mcap_mdl_op_cb *op)
+{
+	if (op->destroy)
+		op->destroy(op->user_data);
+
+	if (op->mdl)
+		mcap_mdl_unref(op->mdl);
+
+	g_free(op);
+}
+
+static void free_mcl_priv_data(struct mcap_mcl *mcl)
+{
+	free_mcap_mdl_op(mcl->priv_data);
+	mcl->priv_data = NULL;
+}
+
 static void mcap_notify_error(struct mcap_mcl *mcl, GError *err)
 {
 	struct mcap_mdl_op_cb *con = mcl->priv_data;
@@ -277,7 +296,7 @@ static void mcap_notify_error(struct mcap_mcl *mcl, GError *err)
 		l = g_slist_find_custom(mcl->mdls, &st, cmp_mdl_state);
 		mdl = l->data;
 		mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-		free_mdl(mdl);
+		mcap_mdl_unref(mdl);
 		update_mcl_state(mcl);
 		con->cb.op_conf(NULL, 0, err, con->user_data);
 		break;
@@ -307,9 +326,7 @@ static void mcap_notify_error(struct mcap_mcl *mcl, GError *err)
 		break;
 	}
 
-	g_free(mcl->priv_data);
-	mcl->priv_data = NULL;
-
+	free_mcl_priv_data(mcl);
 	g_free(mcl->lcmd);
 	mcl->lcmd = NULL;
 }
@@ -440,7 +457,7 @@ static gboolean wait_response_timer(gpointer data)
 	mcap_notify_error(mcl, gerr);
 
 	g_error_free(gerr);
-	mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+	mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 	mcap_cache_mcl(mcl);
 
 	return FALSE;
@@ -451,6 +468,7 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 				uint8_t conf,
 				mcap_mdl_operation_conf_cb connect_cb,
 				gpointer user_data,
+				GDestroyNotify destroy,
 				GError **err)
 {
 	struct mcap_mdl *mdl;
@@ -472,14 +490,15 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 	mdl->state = MDL_WAITING;
 
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op_conf = connect_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 	cmd = create_mdl_req(id, mdepid, conf);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_create_mdl_req),
 									err)) {
-		free_mdl(mdl);
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		g_free(cmd);
 		return FALSE;
@@ -488,7 +507,8 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 	mcl->state = MCL_ACTIVE;
 	mcl->priv_data = con;
 
-	mcl->mdls = g_slist_insert_sorted(mcl->mdls, mdl, compare_mdl);
+	mcl->mdls = g_slist_insert_sorted(mcl->mdls, mcap_mdl_ref(mdl),
+								compare_mdl);
 	mcl->tid = g_timeout_add_seconds(RESPONSE_TIMER, wait_response_timer,
 									mcl);
 	return TRUE;
@@ -497,6 +517,7 @@ gboolean mcap_create_mdl(struct mcap_mcl *mcl,
 gboolean mcap_reconnect_mdl(struct mcap_mdl *mdl,
 				mcap_mdl_operation_cb reconnect_cb,
 				gpointer user_data,
+				GDestroyNotify destroy,
 				GError **err)
 {
 	struct mcap_mdl_op_cb *con;
@@ -508,19 +529,19 @@ gboolean mcap_reconnect_mdl(struct mcap_mdl *mdl,
 					"MDL is not closed");
 		return FALSE;
 	}
-	con = g_new0(struct mcap_mdl_op_cb, 1);
 
 	cmd = create_req(MCAP_MD_RECONNECT_MDL_REQ, mdl->mdlid);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_req), err)) {
-		g_free(con);
 		g_free(cmd);
 		return FALSE;
 	}
 
 	mdl->state = MDL_WAITING;
 
-	con->mdl = mdl;
+	con = g_new0(struct mcap_mdl_op_cb, 1);
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op = reconnect_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 	mcl->state = MCL_ACTIVE;
@@ -553,7 +574,9 @@ static gboolean send_delete_req(struct mcap_mcl *mcl,
 
 gboolean mcap_delete_all_mdls(struct mcap_mcl *mcl,
 					mcap_mdl_notify_cb delete_cb,
-					gpointer user_data, GError **err)
+					gpointer user_data,
+					GDestroyNotify destroy,
+					GError **err)
 {
 	GSList *l;
 	struct mcap_mdl *mdl;
@@ -575,6 +598,7 @@ gboolean mcap_delete_all_mdls(struct mcap_mcl *mcl,
 	con = g_new0(struct mcap_mdl_op_cb, 1);
 	con->mdl = NULL;
 	con->cb.notify = delete_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 
@@ -587,7 +611,9 @@ gboolean mcap_delete_all_mdls(struct mcap_mcl *mcl,
 }
 
 gboolean mcap_delete_mdl(struct mcap_mdl *mdl, mcap_mdl_notify_cb delete_cb,
-					gpointer user_data, GError **err)
+							gpointer user_data,
+							GDestroyNotify destroy,
+							GError **err)
 {
 	struct mcap_mcl *mcl= mdl->mcl;
 	struct mcap_mdl_op_cb *con;
@@ -610,11 +636,13 @@ gboolean mcap_delete_mdl(struct mcap_mdl *mdl, mcap_mdl_notify_cb delete_cb,
 	mdl->state = MDL_DELETING;
 
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.notify = delete_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 	if (!send_delete_req(mcl, con, mdl->mdlid, err)) {
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		return FALSE;
 	}
@@ -623,7 +651,9 @@ gboolean mcap_delete_mdl(struct mcap_mdl *mdl, mcap_mdl_notify_cb delete_cb,
 }
 
 gboolean mcap_mdl_abort(struct mcap_mdl *mdl, mcap_mdl_notify_cb abort_cb,
-					gpointer user_data, GError **err)
+							gpointer user_data,
+							GDestroyNotify destroy,
+							GError **err)
 {
 	struct mcap_mdl_op_cb *con;
 	struct mcap_mcl *mcl = mdl->mcl;
@@ -635,20 +665,19 @@ gboolean mcap_mdl_abort(struct mcap_mdl *mdl, mcap_mdl_notify_cb abort_cb,
 		return FALSE;
 	}
 
-	con = g_new0(struct mcap_mdl_op_cb, 1);
 	cmd = create_req(MCAP_MD_ABORT_MDL_REQ, mdl->mdlid);
 	if (!mcap_send_std_opcode(mcl, cmd, sizeof(mcap_md_req), err)) {
-		g_free(con);
 		g_free(cmd);
 		return FALSE;
 	}
 
-	con->mdl = mdl;
+	con = g_new0(struct mcap_mdl_op_cb, 1);
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.notify = abort_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 	mcl->priv_data = con;
-
 	mcl->tid = g_timeout_add_seconds(RESPONSE_TIMER, wait_response_timer,
 									mcl);
 	return TRUE;
@@ -707,10 +736,8 @@ static void close_mcl(struct mcap_mcl *mcl, gboolean cache_requested)
 		mcl->lcmd = NULL;
 	}
 
-	if (mcl->priv_data) {
-		g_free(mcl->priv_data);
-		mcl->priv_data = NULL;
-	}
+	if (mcl->priv_data)
+		free_mcl_priv_data(mcl);
 
 	g_slist_foreach(mcl->mdls, (GFunc) shutdown_mdl, NULL);
 
@@ -721,7 +748,7 @@ static void close_mcl(struct mcap_mcl *mcl, gboolean cache_requested)
 	if (save)
 		return;
 
-	g_slist_foreach(mcl->mdls, (GFunc) free_mdl, NULL);
+	g_slist_foreach(mcl->mdls, (GFunc) mcap_mdl_unref, NULL);
 	g_slist_free(mcl->mdls);
 	mcl->mdls = NULL;
 }
@@ -745,10 +772,10 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 	if (mcl->ctrl & MCAP_CTRL_CACHED)
 		return;
 
-	mcl->ms->mcls = g_slist_remove(mcl->ms->mcls, mcl);
+	mcl->mi->mcls = g_slist_remove(mcl->mi->mcls, mcl);
 
 	if (mcl->ctrl & MCAP_CTRL_NOCACHE) {
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
 		mcap_mcl_release(mcl);
 		mcap_mcl_unref(mcl);
 		return;
@@ -756,12 +783,12 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 
 	DBG("Caching MCL");
 
-	len = g_slist_length(mcl->ms->cached);
+	len = g_slist_length(mcl->mi->cached);
 	if (len == MAX_CACHED) {
 		/* Remove the latest cached mcl */
-		l = g_slist_last(mcl->ms->cached);
+		l = g_slist_last(mcl->mi->cached);
 		last = l->data;
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, last);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, last);
 		last->ctrl &= ~MCAP_CTRL_CACHED;
 		if (last->ctrl & MCAP_CTRL_CONN) {
 			/* We have to release this MCL if */
@@ -769,12 +796,12 @@ static void mcap_cache_mcl(struct mcap_mcl *mcl)
 			last->ctrl |= MCAP_CTRL_FREE;
 		} else {
 			mcap_mcl_release(last);
-			last->ms->mcl_uncached_cb(last, last->ms->user_data);
+			last->mi->mcl_uncached_cb(last, last->mi->user_data);
 		}
 		mcap_mcl_unref(last);
 	}
 
-	mcl->ms->cached = g_slist_prepend(mcl->ms->cached, mcl);
+	mcl->mi->cached = g_slist_prepend(mcl->mi->cached, mcl);
 	mcl->ctrl |= MCAP_CTRL_CACHED;
 	mcap_mcl_shutdown(mcl);
 }
@@ -786,8 +813,8 @@ static void mcap_uncache_mcl(struct mcap_mcl *mcl)
 
 	DBG("Got MCL from cache");
 
-	mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
-	mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls, mcl);
+	mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
+	mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls, mcl);
 	mcl->ctrl &= ~MCAP_CTRL_CACHED;
 	mcl->ctrl &= ~MCAP_CTRL_FREE;
 }
@@ -813,7 +840,7 @@ void mcap_close_mcl(struct mcap_mcl *mcl, gboolean cache)
 	} else if ((mcl->ctrl & MCAP_CTRL_CACHED) &&
 					(mcl->ctrl & MCAP_CTRL_NOCACHE)) {
 		mcl->ctrl &= ~MCAP_CTRL_CACHED;
-		mcl->ms->cached = g_slist_remove(mcl->ms->cached, mcl);
+		mcl->mi->cached = g_slist_remove(mcl->mi->cached, mcl);
 		mcap_mcl_release(mcl);
 		mcap_mcl_unref(mcl);
 	}
@@ -838,7 +865,7 @@ void mcap_mcl_unref(struct mcap_mcl *mcl)
 		return;
 
 	mcap_mcl_release(mcl);
-
+	mcap_instance_unref(mcl->mi);
 	g_free(mcl->cb);
 	g_free(mcl);
 }
@@ -931,7 +958,7 @@ static void mcap_del_mdl(gpointer elem, gpointer user_data)
 	if (notify)
 		mdl->mcl->cb->mdl_deleted(mdl, mdl->mcl->cb->user_data);
 
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 }
 
 static gboolean check_cmd_req_length(struct mcap_mcl *mcl, void *cmd,
@@ -1020,7 +1047,8 @@ static void process_md_create_mdl_req(struct mcap_mcl *mcl, void *cmd,
 		mdl = g_new0(struct mcap_mdl, 1);
 		mdl->mcl = mcap_mcl_ref(mcl);
 		mdl->mdlid = mdl_id;
-		mcl->mdls = g_slist_insert_sorted(mcl->mdls, mdl, compare_mdl);
+		mcl->mdls = g_slist_insert_sorted(mcl->mdls, mcap_mdl_ref(mdl),
+								compare_mdl);
 	} else if (mdl->state == MDL_CONNECTED) {
 		/* MCAP specification says that we should close the MCL if
 		 * it is open when we receive a MD_CREATE_MDL_REQ */
@@ -1098,6 +1126,7 @@ static void process_md_abort_mdl_req(struct mcap_mcl *mcl, void *cmd,
 	req = cmd;
 	mdl_id = ntohs(req->mdl);
 	mcl->state = MCL_CONNECTED;
+	abrt = NULL;
 	for (l = mcl->mdls; l; l = l->next) {
 		mdl = l->data;
 		if (mdl_id == mdl->mdlid && mdl->state == MDL_WAITING) {
@@ -1308,9 +1337,6 @@ static gboolean process_md_create_mdl_rsp(struct mcap_mcl *mcl,
 	gboolean close;
 	GError *gerr = NULL;
 
-	g_free(mcl->priv_data);
-	mcl->priv_data = NULL;
-
 	close = check_err_rsp(mcl, rsp, len, sizeof(mcap_rsp) + 1, &gerr);
 	g_free(mcl->lcmd);
 	mcl->lcmd = NULL;
@@ -1333,7 +1359,7 @@ static gboolean process_md_create_mdl_rsp(struct mcap_mcl *mcl,
 fail:
 	connect_cb(NULL, 0, gerr, user_data);
 	mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 	g_error_free(gerr);
 	update_mcl_state(mcl);
 	return close;
@@ -1349,9 +1375,6 @@ static gboolean process_md_reconnect_mdl_rsp(struct mcap_mcl *mcl,
 	GError *gerr = NULL;
 	gboolean close;
 
-	g_free(mcl->priv_data);
-	mcl->priv_data = NULL;
-
 	close = check_err_rsp(mcl, rsp, len, sizeof(mcap_rsp), &gerr);
 
 	g_free(mcl->lcmd);
@@ -1372,7 +1395,7 @@ static gboolean process_md_reconnect_mdl_rsp(struct mcap_mcl *mcl,
 	/* Remove cached mdlid */
 	mcl->mdls = g_slist_remove(mcl->mdls, mdl);
 	mcl->cb->mdl_deleted(mdl, mcl->cb->user_data);
-	free_mdl(mdl);
+	mcap_mdl_unref(mdl);
 
 	return close;
 }
@@ -1387,9 +1410,6 @@ static gboolean process_md_abort_mdl_rsp(struct mcap_mcl *mcl,
 	GError *gerr = NULL;
 	gboolean close;
 
-	g_free(mcl->priv_data);
-	mcl->priv_data = NULL;
-
 	close = check_err_rsp(mcl, rsp, len, sizeof(mcap_rsp), &gerr);
 
 	g_free(mcl->lcmd);
@@ -1401,7 +1421,8 @@ static gboolean process_md_abort_mdl_rsp(struct mcap_mcl *mcl,
 
 	if (len >= sizeof(mcap_rsp) && rsp->rc == MCAP_INVALID_MDL) {
 		mcl->mdls = g_slist_remove(mcl->mdls, mdl);
-		free_mdl(mdl);
+		mcl->cb->mdl_deleted(mdl, mcl->cb->user_data);
+		mcap_mdl_unref(mdl);
 	}
 
 	if (gerr)
@@ -1425,6 +1446,18 @@ static void restore_mdl(gpointer elem, gpointer data)
 		mdl->mcl->cb->mdl_closed(mdl, mdl->mcl->cb->user_data);
 }
 
+static void check_mdl_del_err(struct mcap_mdl *mdl, mcap_rsp *rsp)
+{
+	if (rsp->rc != MCAP_ERROR_INVALID_MDL) {
+		restore_mdl(mdl, NULL);
+		return;
+	}
+
+	/* MDL does not exist in remote side, we can delete it */
+	mdl->mcl->mdls = g_slist_remove(mdl->mcl->mdls, mdl);
+	mcap_mdl_unref(mdl);
+}
+
 static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
 								uint32_t len)
 {
@@ -1438,9 +1471,6 @@ static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
 	gboolean close;
 	gboolean notify = FALSE;
 
-	g_free(mcl->priv_data);
-	mcl->priv_data = NULL;
-
 	close = check_err_rsp(mcl, rsp, len, sizeof(mcap_rsp), &gerr);
 
 	g_free(mcl->lcmd);
@@ -1449,7 +1479,7 @@ static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
 
 	if (gerr) {
 		if (mdl)
-			restore_mdl(mdl, NULL);
+			check_mdl_del_err(mdl, rsp);
 		else
 			g_slist_foreach(mcl->mdls, restore_mdl, NULL);
 		deleted_cb(gerr, user_data);
@@ -1473,8 +1503,22 @@ static gboolean process_md_delete_mdl_rsp(struct mcap_mcl *mcl, mcap_rsp *rsp,
 	return close;
 }
 
+static void post_process_rsp(struct mcap_mcl *mcl, struct mcap_mdl_op_cb *op)
+{
+	if (mcl->priv_data != op) {
+		/* Queued MCAP request in some callback. */
+		/* We should not delete the mcl private data */
+		free_mcap_mdl_op(op);
+	} else {
+		/* This is not a queued request. It's safe */
+		/* delete the mcl private data here. */
+		free_mcl_priv_data(mcl);
+	}
+}
+
 static void proc_response(struct mcap_mcl *mcl, void *buf, uint32_t len)
 {
+	struct mcap_mdl_op_cb *op = mcl->priv_data;
 	mcap_rsp *rsp = buf;
 	gboolean close;
 
@@ -1483,15 +1527,19 @@ static void proc_response(struct mcap_mcl *mcl, void *buf, uint32_t len)
 	switch (mcl->lcmd[0] + 1) {
 	case MCAP_MD_CREATE_MDL_RSP:
 		close = process_md_create_mdl_rsp(mcl, rsp, len);
+		post_process_rsp(mcl, op);
 		break;
 	case MCAP_MD_RECONNECT_MDL_RSP:
 		close = process_md_reconnect_mdl_rsp(mcl, rsp, len);
+		post_process_rsp(mcl, op);
 		break;
 	case MCAP_MD_ABORT_MDL_RSP:
 		close = process_md_abort_mdl_rsp(mcl, rsp, len);
+		post_process_rsp(mcl, op);
 		break;
 	case MCAP_MD_DELETE_MDL_RSP:
 		close = process_md_delete_mdl_rsp(mcl, rsp, len);
+		post_process_rsp(mcl, op);
 		break;
 	default:
 		DBG("Unknown cmd response received (op code = %d)", rsp->op);
@@ -1500,7 +1548,7 @@ static void proc_response(struct mcap_mcl *mcl, void *buf, uint32_t len)
 	}
 
 	if (close) {
-		mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 		mcap_cache_mcl(mcl);
 	}
 }
@@ -1591,16 +1639,21 @@ static void mcap_connect_mdl_cb(GIOChannel *chan, GError *conn_err,
 	}
 
 	mdl->state = MDL_CONNECTED;
-	mdl->wid = g_io_add_watch(mdl->dc, G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-						(GIOFunc) mdl_event_cb, mdl);
+	mdl->wid = g_io_add_watch_full(mdl->dc, G_PRIORITY_DEFAULT,
+					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+					(GIOFunc) mdl_event_cb,
+					mcap_mdl_ref(mdl),
+					(GDestroyNotify) mcap_mdl_unref);
 
 	cb(mdl, conn_err, user_data);
 }
 
-gboolean mcap_connect_mdl(struct mcap_mdl *mdl, BtIOType BtType,
+gboolean mcap_connect_mdl(struct mcap_mdl *mdl, uint8_t mode,
 					uint16_t dcpsm,
 					mcap_mdl_operation_cb connect_cb,
-					gpointer user_data, GError **err)
+					gpointer user_data,
+					GDestroyNotify destroy,
+					GError **err)
 {
 	struct mcap_mdl_op_cb *con;
 
@@ -1610,24 +1663,31 @@ gboolean mcap_connect_mdl(struct mcap_mdl *mdl, BtIOType BtType,
 		return FALSE;
 	}
 
+	if ((mode != L2CAP_MODE_ERTM) && (mode != L2CAP_MODE_STREAMING)) {
+		g_set_error(err, MCAP_ERROR, MCAP_ERROR_INVALID_ARGS,
+						"Invalid MDL configuration");
+		return FALSE;
+	}
+
 	con = g_new0(struct mcap_mdl_op_cb, 1);
-	con->mdl = mdl;
+	con->mdl = mcap_mdl_ref(mdl);
 	con->cb.op = connect_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
-	/* TODO: Check if BtIOType is ERTM or Streaming before continue */
-
-	mdl->dc = bt_io_connect(BtType, mcap_connect_mdl_cb, con,
-				g_free, err,
-				BT_IO_OPT_SOURCE_BDADDR, &mdl->mcl->ms->src,
+	mdl->dc = bt_io_connect(BT_IO_L2CAP, mcap_connect_mdl_cb, con,
+				(GDestroyNotify) free_mcap_mdl_op, err,
+				BT_IO_OPT_SOURCE_BDADDR, &mdl->mcl->mi->src,
 				BT_IO_OPT_DEST_BDADDR, &mdl->mcl->addr,
 				BT_IO_OPT_PSM, dcpsm,
 				BT_IO_OPT_MTU, MCAP_DC_MTU,
-				BT_IO_OPT_SEC_LEVEL, mdl->mcl->ms->sec,
+				BT_IO_OPT_SEC_LEVEL, mdl->mcl->mi->sec,
+				BT_IO_OPT_MODE, mode,
 				BT_IO_OPT_INVALID);
 	if (!mdl->dc) {
 		DBG("MDL Connection error");
 		mdl->state = MDL_CLOSED;
+		mcap_mdl_unref(con->mdl);
 		g_free(con);
 		return FALSE;
 	}
@@ -1663,7 +1723,7 @@ fail:
 			mcap_notify_error(mcl, gerr);
 			g_error_free(gerr);
 		}
-		mcl->ms->mcl_disconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_disconnected_cb(mcl, mcl->mi->user_data);
 	}
 	mcap_cache_mcl(mcl);
 	return FALSE;
@@ -1684,7 +1744,7 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 	if (conn_err) {
 		if (mcl->ctrl & MCAP_CTRL_FREE) {
 			mcap_mcl_release(mcl);
-			mcl->ms->mcl_uncached_cb(mcl, mcl->ms->user_data);
+			mcl->mi->mcl_uncached_cb(mcl, mcl->mi->user_data);
 		}
 		connect_cb(NULL, conn_err, data);
 		return;
@@ -1692,7 +1752,7 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 
 	ba2str(&mcl->addr, dstaddr);
 
-	aux = find_mcl(mcl->ms->mcls, &mcl->addr);
+	aux = find_mcl(mcl->mi->mcls, &mcl->addr);
 	if (aux) {
 		/* Double MCL connection case */
 		error("MCL error: Device %s is already connected", dstaddr);
@@ -1714,13 +1774,15 @@ static void mcap_connect_mcl_cb(GIOChannel *chan, GError *conn_err,
 		mcap_uncache_mcl(mcl);
 	else {
 		mcl->ctrl &= ~MCAP_CTRL_FREE;
-		mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls,
+		mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls,
 							mcap_mcl_ref(mcl));
 	}
 
-	mcl->wid = g_io_add_watch(mcl->cc,
+	mcl->wid = g_io_add_watch_full(mcl->cc, G_PRIORITY_DEFAULT,
 				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-				(GIOFunc) mcl_control_cb, mcl);
+				(GIOFunc) mcl_control_cb,
+				mcap_mcl_ref(mcl),
+				(GDestroyNotify) mcap_mcl_unref);
 	connect_cb(mcl, gerr, data);
 }
 
@@ -1730,8 +1792,11 @@ static void set_mdl_properties(GIOChannel *chan, struct mcap_mdl *mdl)
 
 	mdl->state = MDL_CONNECTED;
 	mdl->dc = g_io_channel_ref(chan);
-	mdl->wid = g_io_add_watch(mdl->dc, G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-						(GIOFunc) mdl_event_cb, mdl);
+	mdl->wid = g_io_add_watch_full(mdl->dc, G_PRIORITY_DEFAULT,
+					G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+					(GIOFunc) mdl_event_cb,
+					mcap_mdl_ref(mdl),
+					(GDestroyNotify) mcap_mdl_unref);
 
 	mcl->state = MCL_ACTIVE;
 	mcl->cb->mdl_connected(mdl, mcl->cb->user_data);
@@ -1742,30 +1807,33 @@ static void mcl_io_destroy(gpointer data)
 	struct connect_mcl *con = data;
 
 	mcap_mcl_unref(con->mcl);
+	if (con->destroy)
+		con->destroy(con->user_data);
 	g_free(con);
 }
 
-gboolean mcap_create_mcl(struct mcap_instance *ms,
+gboolean mcap_create_mcl(struct mcap_instance *mi,
 				const bdaddr_t *addr,
 				uint16_t ccpsm,
 				mcap_mcl_connect_cb connect_cb,
 				gpointer user_data,
+				GDestroyNotify destroy,
 				GError **err)
 {
 	struct mcap_mcl *mcl;
 	struct connect_mcl *con;
 
-	mcl = find_mcl(ms->mcls, addr);
+	mcl = find_mcl(mi->mcls, addr);
 	if (mcl) {
 		g_set_error(err, MCAP_ERROR, MCAP_ERROR_ALREADY_EXISTS,
 					"MCL is already connected.");
 		return FALSE;
 	}
 
-	mcl = find_mcl(ms->cached, addr);
+	mcl = find_mcl(mi->cached, addr);
 	if (!mcl) {
 		mcl = g_new0(struct mcap_mcl, 1);
-		mcl->ms = ms;
+		mcl->mi = mcap_instance_ref(mi);
 		mcl->state = MCL_IDLE;
 		bacpy(&mcl->addr, addr);
 		set_default_cb(mcl);
@@ -1777,21 +1845,23 @@ gboolean mcap_create_mcl(struct mcap_instance *ms,
 	con = g_new0(struct connect_mcl, 1);
 	con->mcl = mcap_mcl_ref(mcl);
 	con->connect_cb = connect_cb;
+	con->destroy = destroy;
 	con->user_data = user_data;
 
 	mcl->cc = bt_io_connect(BT_IO_L2CAP, mcap_connect_mcl_cb, con,
 				mcl_io_destroy, err,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_DEST_BDADDR, addr,
 				BT_IO_OPT_PSM, ccpsm,
 				BT_IO_OPT_MTU, MCAP_CC_MTU,
-				BT_IO_OPT_SEC_LEVEL, ms->sec,
+				BT_IO_OPT_SEC_LEVEL, mi->sec,
+				BT_IO_OPT_MODE, L2CAP_MODE_ERTM,
 				BT_IO_OPT_INVALID);
 	if (!mcl->cc) {
 		mcl->ctrl &= ~MCAP_CTRL_CONN;
 		if (mcl->ctrl & MCAP_CTRL_FREE) {
 			mcap_mcl_release(mcl);
-			mcl->ms->mcl_uncached_cb(mcl, mcl->ms->user_data);
+			mcl->mi->mcl_uncached_cb(mcl, mcl->mi->user_data);
 		}
 		mcap_mcl_unref(con->mcl);
 		g_free(con);
@@ -1804,7 +1874,7 @@ gboolean mcap_create_mcl(struct mcap_instance *ms,
 static void connect_dc_event_cb(GIOChannel *chan, GError *gerr,
 							gpointer user_data)
 {
-	struct mcap_instance *ms = user_data;
+	struct mcap_instance *mi = user_data;
 	struct mcap_mcl *mcl;
 	struct mcap_mdl *mdl;
 	GError *err = NULL;
@@ -1823,7 +1893,7 @@ static void connect_dc_event_cb(GIOChannel *chan, GError *gerr,
 		goto drop;
 	}
 
-	mcl = find_mcl(ms->mcls, &dst);
+	mcl = find_mcl(mi->mcls, &dst);
 	if (!mcl || mcl->state != MCL_PENDING)
 		goto drop;
 
@@ -1855,24 +1925,26 @@ static void set_mcl_conf(GIOChannel *chan, struct mcap_mcl *mcl)
 	if (reconn)
 		mcap_uncache_mcl(mcl);
 	else
-		mcl->ms->mcls = g_slist_prepend(mcl->ms->mcls,
+		mcl->mi->mcls = g_slist_prepend(mcl->mi->mcls,
 							mcap_mcl_ref(mcl));
 
-	mcl->wid = g_io_add_watch(mcl->cc,
-			G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-			(GIOFunc) mcl_control_cb, mcl);
+	mcl->wid = g_io_add_watch_full(mcl->cc, G_PRIORITY_DEFAULT,
+				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+				(GIOFunc) mcl_control_cb,
+				mcap_mcl_ref(mcl),
+				(GDestroyNotify) mcap_mcl_unref);
 
 	/* Callback to report new MCL */
 	if (reconn)
-		mcl->ms->mcl_reconnected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_reconnected_cb(mcl, mcl->mi->user_data);
 	else
-		mcl->ms->mcl_connected_cb(mcl, mcl->ms->user_data);
+		mcl->mi->mcl_connected_cb(mcl, mcl->mi->user_data);
 }
 
 static void connect_mcl_event_cb(GIOChannel *chan, GError *gerr,
 							gpointer user_data)
 {
-	struct mcap_instance *ms = user_data;
+	struct mcap_instance *mi = user_data;
 	struct mcap_mcl *mcl;
 	bdaddr_t dst;
 	char address[18], srcstr[18];
@@ -1891,18 +1963,18 @@ static void connect_mcl_event_cb(GIOChannel *chan, GError *gerr,
 		goto drop;
 	}
 
-	ba2str(&ms->src, srcstr);
-	mcl = find_mcl(ms->mcls, &dst);
+	ba2str(&mi->src, srcstr);
+	mcl = find_mcl(mi->mcls, &dst);
 	if (mcl) {
 		error("Control channel already created with %s on adapter %s",
 				address, srcstr);
 		goto drop;
 	}
 
-	mcl = find_mcl(ms->cached, &dst);
+	mcl = find_mcl(mi->cached, &dst);
 	if (!mcl) {
 		mcl = g_new0(struct mcap_mcl, 1);
-		mcl->ms = ms;
+		mcl->mi = mcap_instance_ref(mi);
 		bacpy(&mcl->addr, &dst);
 		set_default_cb(mcl);
 		mcl->next_mdl = (rand() % MCAP_MDLID_FINAL) + 1;
@@ -1927,7 +1999,7 @@ struct mcap_instance *mcap_create_instance(bdaddr_t *src,
 					gpointer user_data,
 					GError **gerr)
 {
-	struct mcap_instance *ms;
+	struct mcap_instance *mi;
 
 	if (sec < BT_IO_SEC_MEDIUM) {
 		g_set_error(gerr, MCAP_ERROR, MCAP_ERROR_INVALID_ARGS,
@@ -1943,54 +2015,55 @@ struct mcap_instance *mcap_create_instance(bdaddr_t *src,
 		return NULL;
 	}
 
-	ms = g_new0(struct mcap_instance, 1);
+	mi = g_new0(struct mcap_instance, 1);
 
-	bacpy(&ms->src, src);
+	bacpy(&mi->src, src);
 
-	ms->sec = sec;
-	ms->mcl_connected_cb = mcl_connected;
-	ms->mcl_reconnected_cb = mcl_reconnected;
-	ms->mcl_disconnected_cb = mcl_disconnected;
-	ms->mcl_uncached_cb = mcl_uncached;
-	ms->mcl_sync_infoind_cb = mcl_sync_info_ind;
-	ms->user_data = user_data;
-	ms->csp_enabled = FALSE;
+	mi->sec = sec;
+	mi->mcl_connected_cb = mcl_connected;
+	mi->mcl_reconnected_cb = mcl_reconnected;
+	mi->mcl_disconnected_cb = mcl_disconnected;
+	mi->mcl_uncached_cb = mcl_uncached;
+	mi->mcl_sync_infoind_cb = mcl_sync_info_ind;
+	mi->user_data = user_data;
+	mi->csp_enabled = FALSE;
 
 	/* Listen incoming connections in control channel */
-	ms->ccio = bt_io_listen(BT_IO_L2CAP, connect_mcl_event_cb, NULL, ms,
+	mi->ccio = bt_io_listen(BT_IO_L2CAP, connect_mcl_event_cb, NULL, mi,
 				NULL, gerr,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_PSM, ccpsm,
 				BT_IO_OPT_MTU, MCAP_CC_MTU,
 				BT_IO_OPT_SEC_LEVEL, sec,
+				BT_IO_OPT_MODE, L2CAP_MODE_ERTM,
 				BT_IO_OPT_INVALID);
-	if (!ms->ccio) {
+	if (!mi->ccio) {
 		error("%s", (*gerr)->message);
-		g_free(ms);
+		g_free(mi);
 		return NULL;
 	}
 
 	/* Listen incoming connections in data channels */
-	ms->dcio = bt_io_listen(BT_IO_L2CAP, connect_dc_event_cb, NULL, ms,
+	mi->dcio = bt_io_listen(BT_IO_L2CAP, connect_dc_event_cb, NULL, mi,
 				NULL, gerr,
-				BT_IO_OPT_SOURCE_BDADDR, &ms->src,
+				BT_IO_OPT_SOURCE_BDADDR, &mi->src,
 				BT_IO_OPT_PSM, dcpsm,
 				BT_IO_OPT_MTU, MCAP_DC_MTU,
 				BT_IO_OPT_SEC_LEVEL, sec,
 				BT_IO_OPT_INVALID);
-	if (!ms->dcio) {
-		g_io_channel_shutdown(ms->ccio, TRUE, NULL);
-		g_io_channel_unref(ms->ccio);
-		ms->ccio = NULL;
+	if (!mi->dcio) {
+		g_io_channel_shutdown(mi->ccio, TRUE, NULL);
+		g_io_channel_unref(mi->ccio);
+		mi->ccio = NULL;
 		error("%s", (*gerr)->message);
-		g_free(ms);
+		g_free(mi);
 		return NULL;
 	}
 
 	/* Initialize random seed to generate mdlids for this instance */
 	srand(time(NULL));
 
-	return ms;
+	return mcap_instance_ref(mi);;
 }
 
 void mcap_release_instance(struct mcap_instance *mi)
@@ -2016,6 +2089,7 @@ void mcap_release_instance(struct mcap_instance *mi)
 		mcap_mcl_release(l->data);
 		mcap_mcl_unref(l->data);
 	}
+
 	g_slist_free(mi->mcls);
 	mi->mcls = NULL;
 
@@ -2023,9 +2097,30 @@ void mcap_release_instance(struct mcap_instance *mi)
 		mcap_mcl_release(l->data);
 		mcap_mcl_unref(l->data);
 	}
+
 	g_slist_free(mi->cached);
 	mi->cached = NULL;
+}
 
+struct mcap_instance *mcap_instance_ref(struct mcap_instance *mi)
+{
+	mi->ref++;
+
+	DBG("mcap_instance_ref(%p): ref=%d", mi, mi->ref);
+
+	return mi;
+}
+
+void mcap_instance_unref(struct mcap_instance *mi)
+{
+	mi->ref--;
+
+	DBG("mcap_instance_unref(%p): ref=%d", mi, mi->ref);
+
+	if (mi->ref > 0)
+		return;
+
+	mcap_release_instance(mi);
 	g_free(mi);
 }
 
@@ -2064,3 +2159,37 @@ uint16_t mcap_get_data_psm(struct mcap_instance *mi, GError **err)
 
 	return lpsm;
 }
+
+gboolean mcap_set_data_chan_mode(struct mcap_instance *mi, uint8_t mode,
+								GError **err)
+{
+	if (!(mi && mi->dcio)) {
+		g_set_error(err, MCAP_ERROR, MCAP_ERROR_INVALID_ARGS,
+						"Invalid MCAP instance");
+		return FALSE;
+	}
+
+	return bt_io_set(mi->dcio, BT_IO_L2CAP, err, BT_IO_OPT_MODE, mode,
+							BT_IO_OPT_INVALID);
+}
+
+struct mcap_mdl *mcap_mdl_ref(struct mcap_mdl *mdl)
+{
+	mdl->ref++;
+
+	DBG("mcap_mdl_ref(%p): ref=%d", mdl, mdl->ref);
+
+	return mdl;
+}
+
+void mcap_mdl_unref(struct mcap_mdl *mdl)
+{
+	mdl->ref--;
+
+	DBG("mcap_mdl_unref(%p): ref=%d", mdl, mdl->ref);
+
+	if (mdl->ref > 0)
+		return;
+
+	free_mdl(mdl);
+}
