@@ -260,6 +260,7 @@ static int pending_remote_name_cancel(struct btd_adapter *adapter)
 	if (!dev) /* no pending request */
 		return -ENODATA;
 
+	adapter->state &= ~STATE_RESOLVNAME;
 	err = adapter_ops->cancel_resolve_name(adapter->dev_id, &dev->bdaddr);
 	if (err < 0)
 		error("Remote name cancel failed: %s(%d)",
@@ -1024,6 +1025,9 @@ static void adapter_emit_uuids_updated(struct btd_adapter *adapter)
 	int i;
 	sdp_list_t *list;
 
+	if (!adapter->initialized)
+		return;
+
 	uuids = g_new0(char *, sdp_list_len(adapter->services) + 1);
 
 	for (i = 0, list = adapter->services; list; list = list->next) {
@@ -2401,24 +2405,19 @@ void btd_adapter_get_mode(struct btd_adapter *adapter, uint8_t *mode,
 	if (mode) {
 		if (main_opts.remember_powered == FALSE)
 			*mode = main_opts.mode;
-		else if (read_device_mode(address, str, sizeof(str)) < 0)
-			*mode = main_opts.mode;
-		else
+		else if (read_device_mode(address, str, sizeof(str)) == 0)
 			*mode = get_mode(&adapter->bdaddr, str);
+		else
+			*mode = main_opts.mode;
 	}
 
 	if (on_mode) {
-		if (main_opts.remember_powered == FALSE) {
-			if (adapter->initialized)
-				*on_mode = get_mode(&adapter->bdaddr, "on");
-			else {
-				*on_mode = main_opts.mode;
-				adapter->initialized = TRUE;
-			}
-		} else if (read_on_mode(address, str, sizeof(str)) < 0)
-			*on_mode = main_opts.mode;
-		else
+		if (main_opts.remember_powered == FALSE)
+			*on_mode = get_mode(&adapter->bdaddr, "on");
+		else if (read_on_mode(address, str, sizeof(str)) == 0)
 			*on_mode = get_mode(&adapter->bdaddr, str);
+		else
+			*on_mode = main_opts.mode;
 	}
 
 	if (pairable)
@@ -2506,6 +2505,15 @@ static void set_mode_complete(struct btd_adapter *adapter)
 
 	DBG("");
 
+	/*
+	 * g_slist_free is not called after g_slist_foreach because the list is
+	 * updated using g_slist_remove in session_remove which is called by
+         * session_free, which is called for each element by g_slist_foreach.
+	 */
+	if (adapter->mode == MODE_OFF)
+		g_slist_foreach(adapter->mode_sessions, (GFunc) session_free,
+									NULL);
+
 	if (adapter->pending_mode == NULL)
 		return;
 
@@ -2722,6 +2730,8 @@ gboolean adapter_init(struct btd_adapter *adapter)
 	 * the are active connections before the daemon've started */
 	load_connections(adapter);
 
+	adapter->initialized = TRUE;
+
 	return TRUE;
 }
 
@@ -2817,12 +2827,15 @@ void adapter_set_state(struct btd_adapter *adapter, int state)
 		break;
 	case STATE_LE_SCAN:
 		/* Scanning enabled */
-		adapter->stop_discov_id = g_timeout_add(5120,
-						stop_scanning, adapter);
+		if (adapter->disc_sessions) {
+			adapter->stop_discov_id = g_timeout_add(5120,
+								stop_scanning,
+								adapter);
 
-		/* For dual mode: don't send "Discovering = TRUE"  */
-		if (bredr_capable(adapter) == TRUE)
-			return;
+			/* For dual mode: don't send "Discovering = TRUE"  */
+			if (bredr_capable(adapter) == TRUE)
+				return;
+		}
 
 		/* LE only */
 		discov_active = TRUE;
@@ -3719,3 +3732,21 @@ int adapter_cancel_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 {
 	return adapter_ops->cancel_bonding(adapter->dev_id, bdaddr);
 }
+
+int btd_adapter_read_local_oob_data(struct btd_adapter *adapter)
+{
+	return adapter_ops->read_local_oob_data(adapter->dev_id);
+}
+
+int btd_adapter_add_remote_oob_data(struct btd_adapter *adapter,
+			bdaddr_t *bdaddr, uint8_t *hash, uint8_t *randomizer)
+{
+	return adapter_ops->add_remote_oob_data(adapter->dev_id, bdaddr, hash,
+								randomizer);
+}
+
+int btd_adapter_remove_remote_oob_data(struct btd_adapter *adapter,
+							bdaddr_t *bdaddr)
+{
+	return adapter_ops->remove_remote_oob_data(adapter->dev_id, bdaddr);
+}
